#lang racket
(provide make-withdraw make-account send look-up-method)

;; ---------------------------------------------------------------------
;; our factory for withdraw procedures, from last time
;; ---------------------------------------------------------------------

(define make-withdraw
  (lambda (balance)
    ; --------------
    (lambda (amount)                 ;; balance is free inside this
      (if (>= balance amount)        ;; lambda but bound in context:
          (begin                     ;; to a new object on each call!
            (set! balance (- balance amount))
            balance)
          (error "Insufficient funds" balance)))
    ; --------------
  ))

;; ---------------------------------------------------------------------
;; a more full-featured account, with withdraw and deposit behaviors
;; ---------------------------------------------------------------------

(define make-account        ;; make-account is a procedure that creates
  (lambda (balance)         ;; a closure around balance.

    (lambda (transaction)      ;; The procedure it returns uses its arg
      (case transaction        ;; to select a particular operation on
        ('withdraw             ;; balance  --  returning a procedure!
           (lambda (amount)       
             (if (>= balance amount)
                 (begin
                   (set! balance (- balance amount))
                   balance)
                 (error "Insufficient funds" balance))))
        ('deposit
           (lambda (amount)
             (set! balance (+ balance amount))
             balance))
        (else
           (error "Unknown request to account:" transaction)))) ))

;; ---------------------------------------------------------------------
;; syntactic sugar for interacting with our closure "objects"
;; ---------------------------------------------------------------------

(define send
  (lambda (object message . args)
    (apply (look-up-method object message) args)))

(define look-up-method
  (lambda (object selector)
    (object selector)))

;; does look-up-method remind you of apply-ff?
;; ---------------------------------------------------------------------
