#lang racket
(provide make-account withdraw deposit)

;; ---------------------------------------------------------------------
;; a list of procedures that share a common variable
;; ---------------------------------------------------------------------

(define make-account        ;; Create a procedure that creates
  (lambda (balance)         ;; a closure around balance.
    (list (lambda (amount)
            (if (>= balance amount)
                (begin (set! balance (- balance amount))
                       balance)
               (error "Insufficient funds" balance)))
          (lambda (amount)
            (set! balance (+ balance amount))
            balance))))

(define withdraw first)
(define deposit  second)

;; ---------------------------------------------------------------------
;;
;; > (define account-for-eugene (make-account 100))
;; > ((withdraw account-for-eugene) 10)
;; 90
;; > ((deposit account-for-eugene) 10)
;; 100
;;
;; ---------------------------------------------------------------------
;;
;; Can we define withdraw and deposit in a way that doesn't require
;; the extra parentheses?
;;
;; > (withdraw account-for-eugene 10)
;; > (deposit account-for-eugene 10)
;;
;; Sure:
;;
;; (define withdraw
;;   (lambda (object . args)
;;     (apply (first object) args)))
;;
;; ---------------------------------------------------------------------

;; ---------------------------------------------------------------------
;; We can also adapt our "send" function to hide the new list-based way
;; of interating with the objects.
;; ---------------------------------------------------------------------

(define send                        ; this stays the same...
  (lambda (object message . args)
    (apply (look-up-method object message) args)))

(define look-up-method              ; ... but this changes
  (lambda (object selector)
    (let ((position (list-index selector '(withdraw deposit))))
      (list-ref object position))))

;; ---------------------------------------------------------------------
;; an old friend
;; ---------------------------------------------------------------------

(define list-index
  (lambda (target los)
    (letrec ((list-index-helper
               (lambda (los base)
                 (if (null? los)
                     -1
                     (if (eq? target (first los))
                         base
                         (list-index-helper (rest los)
                                            (add1 base)))))))
      (list-index-helper los 0))))

;; ---------------------------------------------------------------------

