#lang racket
(provide make-account)

;; ----------------------------------------------------------------------
;; Another way to implement message passing.  Our method selector
;; is created in a let that defines the methods as local functions.
;;
;; Note: I changed the variable name to "acct-balance" so that we can
;; use "balance" as the name of a local function.
;; ----------------------------------------------------------------------

(define make-account        ;; Creates a procedure that creates
  (lambda (acct-balance)    ;; a closure around acct-balance.
    (let ((withdraw               ;; These procedures share
            (lambda (amount)      ;; the balance variable.
              (if (>= acct-balance amount)
                  (begin
                    (set! acct-balance (- acct-balance amount))
                    acct-balance)
                  (error "Insufficient funds" acct-balance))))
          (deposit
            (lambda (amount)
              (set! acct-balance (+ acct-balance amount))
              acct-balance))
          (balance
            (lambda ()
              acct-balance))
          (error
            (lambda args
              (error "Unknown request -- ACCOUNT"))))
      (lambda (transaction)   
        (case transaction     
          ('withdraw withdraw)
          ('deposit  deposit)
          ('balance  balance)
          (else      error))))))

;; ----------------------------------------------------------------------

