#lang racket
(provide make-account)

;; ----------------------------------------------------------------------
;; This account object responds to messages on its own!
;;
;; Note: I changed the variable name to "acct-balance" so that we can
;; use "balance" as the name of a local function.
;; ----------------------------------------------------------------------

(define make-account        ;; Defines a procedure that creates
  (lambda (acct-balance)    ;; a closure around acct-balance.
    (let ((withdraw               ;; These procedures share
            (lambda (amount)      ;; the balance variable.
              (if (>= acct-balance amount)
                  (begin
                    (set! acct-balance (- acct-balance amount))
                    acct-balance)
                  (error "Insufficient funds" acct-balance))))
          (deposit
            (lambda (amount)
              (set! acct-balance (+ acct-balance amount))
              acct-balance))
          (balance
            (lambda ()
              acct-balance)))
      (lambda (transaction . args)       ;; This procedure *applies*
        (case transaction                ;; the "method" corresponding
          ('withdraw                     ;; to the transaction to its
             (apply withdraw args))      ;; arguments.
          ('deposit
             (apply deposit  args))      ;; How can we make the repeated
          ('balance                      ;; (apply <selector> args)
             (apply balance  args))      ;; go away?
          (else
             (error "Unknown request -- ACCOUNT" transaction)))))))

;; ----------------------------------------------------------------------

